/* ------------------
 * Logic2CNF
 * $Id: cnf_clauses.c 509 2010-06-11 10:29:55Z eb771 $
 *
 * Copyright (c) 2010, Edd Barrett (University of Kent)
 * <e.barrett@kent.ac.uk>
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "logic2cnf.h"
#include "y.tab.h"
#include "queue.h"

/* from lits.c */
extern RB_HEAD(cnf_tree, literal_)	literals;

TAILQ_HEAD(ncs_head, cnf_clause_)	cnf_clauses;
unsigned long int			num_cnf_clauses = 0;
cnf_clause				*cur_cnf_clause;

/*
 * flatten a CNF tree into lists of disjunctions and
 * store in global 'cnf_clauses'.
 */
void
make_cnf_clauses(lnode *node, int *clause_ct)
{
	int	ct_left = 0, ct_right = 0;

	switch(node->type)  {
		case AND:
			assert(node->right != NULL);
			assert(node->left != NULL);

			make_cnf_clauses(node->left, &ct_left);
			if (ct_left > 0)
				insert_cur_cnf_clause();

			make_cnf_clauses(node->right, &ct_right);
			if (ct_right > 0)
				insert_cur_cnf_clause();

			break;
		case OR:
			assert(node->right != NULL);
			assert(node->left != NULL);

			make_cnf_clauses(node->left, clause_ct);
			make_cnf_clauses(node->right, clause_ct);
			break;
		case NOT: /* NOT will alway be followed by a LIT in CNF */
			assert(node->left->type == LIT);
			/* FALLTHRU */
		case LIT:
			add_lit_to_cur_cnf_clause(node);
			(*clause_ct) ++;

			break;
		default:
			lerr(1, "make_cnf_clauses():\n\tunknown node type");
	}
}

/* insert a node clause, expand if necessary */
void
insert_cur_cnf_clause()
{
	cur_cnf_clause->dead = 0;
	cur_cnf_clause->deduped = 0;
	TAILQ_INSERT_TAIL(&cnf_clauses, cur_cnf_clause, entries);
	num_cnf_clauses ++;
	cur_cnf_clause = NULL; /* mark for creation */
}

/* insert a literal into current clause */
void
add_lit_to_cur_cnf_clause(lnode *node)
{
	cnf_clause_lit		*item;
	uint8_t			negated;
	lnode			*subject;

	/* if not yet added any clauses, make space for a list */
	if (cur_cnf_clause == NULL) {
		cur_cnf_clause = (cnf_clause *)
		    xmalloc(sizeof(cnf_clause)); 
		TAILQ_INIT(&cur_cnf_clause->lits);
		cur_cnf_clause->len = 0;
	}

	/* try to make the only place you call this */
	subject = get_node_subject(node, &negated);

	item = (cnf_clause_lit *) xmalloc(sizeof(cnf_clause_lit));

	item->lit = subject->lit;
	item->negated = negated;

	cur_cnf_clause->len ++;
	TAILQ_INSERT_HEAD(&(cur_cnf_clause->lits), item, entries);
}


/*
 * free the node clause lists, but NOT the nodes themselves, they are
 * pool allocated in all_nodes and are freed later.
 */
void
free_cnf_clauses()
{
	cnf_clause	*clause, *next;

	for (clause = TAILQ_FIRST(&cnf_clauses);
	    clause != TAILQ_END(&cnf_clauses);
	    clause = next) {
		next = TAILQ_NEXT(clause, entries);
		free_cnf_clause(clause);
	}
}

/* free list bits for each node clause */
void
free_cnf_clause(cnf_clause *clause)
{
	cnf_clause_lit		*lit, *next;

	for (lit = TAILQ_FIRST(&clause->lits);
	    lit != TAILQ_END(&clause->lits);
	    lit = next) {
		next = TAILQ_NEXT(lit, entries);
		free(lit);
	}
	free(clause);
}

void
remove_dead_cnf_clauses()
{
	unsigned long int	rm_ct = 0;
	cnf_clause		*clause, *next;

	for (clause = TAILQ_FIRST(&cnf_clauses);
	    clause != TAILQ_END(&cnf_clauses);
	    clause = next) {
		next = TAILQ_NEXT(clause, entries);

		if (clause->dead == 1) {
			TAILQ_REMOVE(&cnf_clauses, clause, entries);
			free_cnf_clause(clause);
			num_cnf_clauses --;
			rm_ct ++;
		} else {
			clause->deduped = 0;
		}
	}

	if (l_debug)
		fprintf(stderr, "This time we removed %lu clauses\n", rm_ct);
}

void
mark_dead_trivially_true_cnf_clauses()
{
	cnf_clause		*item;

	TAILQ_FOREACH(item, &cnf_clauses, entries) {
		if (cnf_clause_is_trivially_true(item)) {
			if (l_debug) {
				fprintf(stderr,
				    "removing trivially true clause: ");
				cnf_clause_to_text(item, 1);
				fprintf(stderr, "\n");
			}
			item->dead = 1;
		}
	}
}

/*
 * Often what we get is identical node clauses, we want to 
 * get rid of those ASAP as they cost time to scan. 
 */
void
mark_identical_cnf_clauses_dead()
{
	cnf_clause			*clause1, *clause2;

	TAILQ_FOREACH(clause1, &cnf_clauses, entries) {

		if ((clause1->deduped) || (clause1->dead))
			continue;

		clause1->deduped = 1;
		TAILQ_FOREACH(clause2, &cnf_clauses, entries) {

			if ((clause2->deduped) || (clause2->dead))
				continue;

			if (cnf_clauses_are_identical(clause1, clause2)) {
				clause2->dead = 1;
				clause2->deduped = 1;
			}
		}
	}
}


/* find a lit with matching negation in a node clause */
cnf_clause_lit *
find_lit_in_cnf_clause(cnf_clause_lit *cl_lit, cnf_clause *clause)
{
	cnf_clause_lit			*item;

	TAILQ_FOREACH(item, &clause->lits, entries) {
		if ((item->lit == cl_lit->lit)
		    && (item->negated == cl_lit->negated))
			return item;
	}
	return NULL;
}

/* just find any reference to a literal in a node clause */
cnf_clause_lit *
find_lit_in_cnf_clause(literal *lit, cnf_clause *clause)
{
	cnf_clause_lit		*l;

	TAILQ_FOREACH(l, &clause->lits, entries)
		if (l->lit == lit)
			return l;

	return 0;
}

/* similar to the above, but for *all* node clauses */
cnf_clause_lit *
find_lit_in_cnf_clauses(literal *lit)
{
	cnf_clause             *clause;
	cnf_clause_lit		*cl_lit;
	uint8_t                 found_lit;

	found_lit = 0;
	TAILQ_FOREACH(clause, &cnf_clauses, entries) {
		cl_lit = find_lit_in_cnf_clause(lit, clause);
		if (cl_lit)
			return cl_lit;
	}
	return 0;
}

void
cnf_clause_to_dimacs(cnf_clause *clause)
{
	cnf_clause_lit		*item;

	TAILQ_FOREACH(item, &clause->lits, entries) {
		if (!item->negated)
			printf("%lu ", item->lit->cnf_input_num);
		else
			printf("-%lu ", item->lit->cnf_input_num);
	}
	printf("0\n");
}

void
all_cnf_clauses_to_dimacs()
{
	cnf_clause	*clause;

	TAILQ_FOREACH(clause, &cnf_clauses, entries)
		cnf_clause_to_dimacs(clause);
}

/* print a dimacs header */
void
dimacs_header()
{
	literal			*item;

	printf("c auto-generated by logic2cnf "
	    PACKAGE_VERSION " (Build: " __DATE__ " " __TIME__
	    ")\nc +---------------\n");

	RB_FOREACH(item, cnf_tree, &literals) {
		if (resolve_temps && item->internal_witness)
			continue;

		printf("c | %6lu = %s\n",
		    item->cnf_input_num, item->name);
	}
	printf("\n");
}

/*
 * print out the logical clause in a textual format.
 * For example: (a b ~c)
 */
void
cnf_clause_to_text(cnf_clause *clause, uint8_t to_stderr)
{
	cnf_clause_lit		*item;
	uint8_t			first = 1;
	FILE			*where = stdout;

	if (to_stderr)
		where = stderr;

	fprintf(where, "( ");
	TAILQ_FOREACH(item, &clause->lits, entries) {

		if (!first)
			fprintf(where, "+ ");

		if (!item->negated)
			fprintf(where, "%s ", item->lit->name);
		else
			fprintf(where, "~%s ", item->lit->name);

		first = 0;
	}
	fprintf(where, ")");
}

/* 
 * print *all* clauses in textual format
 */
void
all_cnf_clauses_to_text(uint8_t to_stderr)
{
	cnf_clause			*clause;
	literal				*lit;
	unsigned long int		i;
	FILE				*where = stdout;

	if (to_stderr)
		where = stderr;

	/* print a header */
	fprintf(where, "# ----------------------------------------\n");
	fprintf(where, "# CNF generated by logic2nf: " PACKAGE_VERSION "\n");
	fprintf(where, "# ----------------------------------------\n\n");

	/* print definitions for all literals */ 
	RB_FOREACH(lit, cnf_tree, &literals) {

		/* temp lits will have gone if optimiser on */
		if (resolve_temps  &&
		    (lit->internal_witness))
			continue;

		fprintf(where, "def %s;\n", lit->name); 
	} 
	fprintf(where, "\n"); 

	i = 1;
	TAILQ_FOREACH(clause, &cnf_clauses, entries) {
		cnf_clause_to_text(clause, to_stderr);

		if (i++ != (num_cnf_clauses))
			fprintf(where, " .");

		fprintf(where, "\n");
	}

	fprintf(where, ";\n");
}

/* ---[ MISC ]------------------------------------------------------------ */

uint8_t
cnf_clause_is_trivially_true(cnf_clause *clause)
{
	cnf_clause_lit			*item1, *item2;

	TAILQ_FOREACH(item1, &clause->lits, entries) {
		TAILQ_FOREACH(item2, &clause->lits, entries) {

			/* dont compare selves */
			if (item1 == item2)
				continue;

			if ((item1->lit == item2->lit) &&
			    (item1->negated != item2->negated))
				return 1; /* yup, trivial */
		}
	}
	return 0;
}

uint8_t
cnf_clauses_are_identical(cnf_clause *clause1, cnf_clause *clause2)
{
	cnf_clause_lit		*lit;

	if (clause1 == clause2)
		return 1;

	if (clause1->len != clause2->len)
		return 0;

	/*
	 * ok, so we have two clauses of same length
	 * that are *not* the same instance, now are they
	 * duplicates of each other?
	 */	
	TAILQ_FOREACH(lit, &clause1->lits, entries) {
		if (!find_lit_in_cnf_clause(lit, clause2))
			return 0;
	}

	if (l_debug) {
		fprintf(stderr, "removing duplicate clause: ");
		cnf_clause_to_text(clause1, 1);
		fprintf(stderr, "\n");
	}

	return 1;
}

void
remove_duplicate_literals_from_cnf_clauses()
{
	cnf_clause		*cl;

	TAILQ_FOREACH(cl, &cnf_clauses, entries) {
		remove_duplicate_literals_from_cnf_clause(cl);
	}
}

void
remove_duplicate_literals_from_cnf_clause(cnf_clause *clause)
{
	cnf_clause_lit		*pri_lit;
	cnf_clause_lit		*sec_lit, *sec_next;

	TAILQ_FOREACH(pri_lit, &clause->lits, entries) {

		for (sec_lit = TAILQ_FIRST(&clause->lits);
		    sec_lit != TAILQ_END(&clause->nodes);
		    sec_lit = sec_next) {
			sec_next = TAILQ_NEXT(sec_lit, entries);

			/* that would suck, don't dedupe against self */
			if (pri_lit == sec_lit)
				continue;

			if ((pri_lit->lit == sec_lit->lit)
			    && (pri_lit->negated == sec_lit->negated)) {
				/* duplicate found */

				if (l_debug) {
					fprintf(stderr, "removing duplicate "
					    "literal '%s' from clause: ",
					    pri_lit->lit->name);

					cnf_clause_to_text(clause, 1);
					fprintf(stderr, "\n");
				}
				TAILQ_REMOVE(&clause->lits, sec_lit, entries);
				clause->len --;
				free(sec_lit);
			}
		}
	}
}

/*
 * wash, scrub, rinse and bleach CNF clauses.
 * this is SLOW, so call only if the user turns the optimiser on.
 */
void
clean_cnf_clauses() {
	mark_dead_trivially_true_cnf_clauses();
	mark_identical_cnf_clauses_dead();
	mark_supersets_dead();
	remove_dead_cnf_clauses();
}
