/* ------------------
 * Logic2CNF
 * $Id: graphviz.c 504 2010-06-10 15:01:13Z eb771 $
 *
 * Copyright (c) 2010, Edd Barrett (University of Kent)
 * <e.barrett@kent.ac.uk>
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <stdio.h>
#include <stdlib.h>

#include "logic2cnf.h"
#include "y.tab.h"

FILE			*gv_out = NULL;
unsigned long int	gv_id = 0;

/*
 * Open graphviz output file and write digraph header
 */
void
gv_init(const char *filename)
{
	assert(gv_out == NULL);

	gv_out = fopen(filename, "w");
	if (gv_out == NULL) {
		perror("gv_init(): cannot fopen()");
		goto fail;
	}

	/* header */
	if (fprintf(gv_out, "digraph G {\n") == -1) {
		perror("gv_init(): cannot fprintf");
		goto fail;
	}

	return;

fail:
	free_everything();
	exit(EXIT_FAILURE);
}

/*
 * Write graphviz footer and close file
 */
void
gv_done()
{
	if (fprintf(gv_out, "\n}") == -1) {
		perror("gv_done(): cannot fprintf()");
		goto fail;
	}

	fclose(gv_out);
	gv_out = NULL;

	return;

fail:
	free_everything();
	exit(EXIT_FAILURE);
}

/*
 * Draw tree nodes into a graphviz file
 */
unsigned long int
gv_draw_tree(lnode *node)
{
	char			*node_str = NULL;
	unsigned long int	child_gv_id, my_gv_id;

	assert(gv_out != NULL);

	node_str = get_node_str(node);
	my_gv_id = gv_id ++;
	
	switch (node->type) {
		case LIT:
			fprintf(gv_out, "\t%lu [label=\"%s\"];\n",
			    my_gv_id, node_str);
		break;

		case AND:
		case OR:
		case NOT:
			if (fprintf(gv_out, "\t%lu [label=\"%s\"];\n",
			    my_gv_id, node_str) == -1) {
				perror("gv_draw_tree(): cannot fprintf()");
				goto fail;
			}

			/* recurse left */
			child_gv_id = gv_draw_tree(node->left);

			if (fprintf(gv_out, "\t%lu -> %lu;\n",
			    my_gv_id, child_gv_id) == -1) {
				perror("gv_draw_tree(): cannot fprintf()");
				goto fail;
			}

			/* recurse right */
			if (node->right != NULL) {

				child_gv_id = gv_draw_tree(node->right);

				if (fprintf(gv_out, "\t%lu -> %lu;\n",
				    my_gv_id, child_gv_id) == -1) {
					perror("gv_draw_tree(): "
					    "cannot fprintf()");
					goto fail;
				}
			}

		break;
		default:
			fprintf(stderr, "gv_draw_tree(): unknown node type");
		goto fail;
	};

	free(node_str);
	return my_gv_id;

fail:
	free(node_str);
	free_everything();
	exit(EXIT_FAILURE);
}
