/* ------------------
 * Logic2CNF
 * $Id: lits.c 505 2010-06-10 15:10:00Z eb771 $
 *
 * Copyright (c) 2010, Edd Barrett (University of Kent)
 * <e.barrett@kent.ac.uk>
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include "logic2cnf.h"
#include "y.tab.h"
#include "queue.h"

unsigned long int 			num_literals = 0;
unsigned long int 			num_witness = 0;

/* tree comparison function */
int
cnf_cmp(literal *c1, literal *c2)
{
	return strcmp(c1->name, c2->name);
}

/* tree itself */
RB_HEAD(cnf_tree, literal_)	literals = RB_INITIALIZER(&literals);
RB_GENERATE(cnf_tree, literal_, entries, cnf_cmp)

/* lookup a literals CNF input number from it's name */
literal *
get_literal(char *litname)
{
	literal		find;

	if (litname == NULL)
		return NULL;

	find.name = litname;
	return RB_FIND(cnf_tree, &literals, &find);
}

/* define a literal in the symbol table */
literal *
define_lit(char *name, uint8_t internal)
{
	literal		*cnf_item = get_literal(name);

	/* check literal not in use */
	if (cnf_item != NULL) {
		lerr(0, "%s line %d:\n\tliteral '%s' redefined\n"
		    "\tliteral '%s' first defined on line %d",
		    yyin_filename, my_yylineno, name,
		    name, cnf_item->line_defined);
		num_symbol_errors ++;
	} else {
		/* populate info */
		cnf_item = (literal *)
		    xmalloc(sizeof(literal));

		cnf_item->name = strdup(name); 
		cnf_item->used = 0;
		cnf_item->line_defined = my_yylineno;
		cnf_item->internal_witness = internal;
		cnf_item->minisat_lit = NULL; /* made later */

		RB_INSERT(cnf_tree, &literals, cnf_item);
		num_literals ++;
	}
	return cnf_item;
}

literal *
get_literal_by_num(unsigned long int input)
{
	literal		*item;

	assert(input > 0);

	RB_FOREACH(item, cnf_tree, &literals)
		if (item->cnf_input_num == input)
			return item;

	return NULL; /* not found */
}

/* 
 * rename _x lits to tmp__x (for use when resollving -l output)
 * If we did not do this, then we could not re-solve as the formula
 * would use reserved literal prefixes _x.
 */
void
rename_internal_witness_lits()
{
	char			*old_name;
	literal		*item;

	/* scan for _x literals */
	RB_FOREACH(item, cnf_tree, &literals) {

		if (! item->internal_witness)
			continue;

		old_name = strdup(item->name);
		
		/* rename lit, by prefixing tmp_ */
		free(item->name);
		if (asprintf(&item->name, "tmp_%s", old_name) == -1)
			lerr(1,
			    "rename_internal_witness_lits():\n\t"
			    "cannot asprintf(): %s", strerror(errno));

		free(old_name);
	}
}

void
free_literals()
{
	literal		*cnf_item, *next_cnf_item;

	for (cnf_item = RB_MIN(cnf_tree, &literals);
	    cnf_item != NULL;
	    cnf_item = next_cnf_item) {
		next_cnf_item = RB_NEXT(cnf_tree, &literals, cnf_item);
		RB_REMOVE(cnf_tree, &literals, cnf_item);
		free_literal(cnf_item);
	}
	num_literals = 0;
}

void
free_literal(literal *input)
{
	free(input->name);
	if (input->minisat_lit != NULL)
		delete(input->minisat_lit);

	free(input);
}
