/* ------------------
 * Logic2CNF
 * $Id: logic2cnf.c 515 2010-06-11 11:15:11Z eb771 $
 *
 * Copyright (c) 2010, Edd Barrett (University of Kent)
 * <e.barrett@kent.ac.uk>
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/* linux libc needs this for asprintf() */
#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <assert.h>
#include <fcntl.h>
#include <stdarg.h>

/* all !CXX! */
#include <iostream>
#include <set>
#include <Solver.h>
#include <SolverTypes.h>

using namespace std;

#include <sys/types.h>
#include <sys/wait.h>

#include "queue.h"
#include "logic2cnf.h"
#include "y.tab.h"
#include "config.h"

/* from nodes.c */
extern SIMPLEQ_HEAD(alln_head, lnode_)		all_nodes;

/* from lits.c */
extern RB_HEAD(cnf_tree, literal_)		literals;

/* from cnf_clauses.c */
extern TAILQ_HEAD(ncs_head, cnf_clause_)	cnf_clauses;

/* from sat.c */
extern TAILQ_HEAD(sr_head, sat_result_)	sat_results;
extern SIMPLEQ_HEAD(srp_head, sat_report_)	sat_reports;

/* from uq_elims.c */
extern SLIST_HEAD(uqe_head, uq_elim_lit_)	uq_elims;

/* from resolve.c */
extern SLIST_HEAD(rsl_head, resolve_lit_)	resolve_lits;

int			num_symbol_errors = 0;
unsigned long int	indent = 0;	
uint8_t			l_debug = 0;
uint8_t			resolve_temps = 0;
uint8_t			enable_progress_report = 1; /* -p to disable */
char			*yyin_filename = NULL;

/* 
 * dump a tree of lnodes to stderr following thier left/right
 * node pointers
 */
void
dump_tree(lnode *node) {
	unsigned long int 	i;
	char			*node_str = get_node_str(node);

	/* indent node appropriately */
	for (i = 0; i < indent; i ++)
		fprintf(stderr, "  ");
	indent ++;

	fprintf(stderr, "[%s]\n", node_str);
	free(node_str);

	/* traverse children if they exist */
	if (node->left != NULL)
		dump_tree(node->left);

	if (node->right != NULL)
		dump_tree(node->right);
	indent --;
}

/* 
 * Execute M4 on the input file and save to a temp file.
 * Return the new filename.
 */
char
*invoke_m4(char *filename)
{
	char	*new_filename;
	int	pre_fd = 0, post_fd, pid;
	int	status = -1;

	new_filename = strdup(M4_TMP_TEMPLATE);
	if (new_filename == NULL)
		lerr(1, "invoke_m4():\n\tcannot strdup(): %s",
		    strerror(errno));

	post_fd = mkstemp(new_filename);
	if (post_fd == -1)
		lerr(1, "invoke_m4():\n\tcannot mkstemp(): %s",
		    strerror(errno));

	pid = fork();
	if (!pid) { /* child */

		if (filename != NULL) {
			pre_fd = open(filename, O_RDONLY, NULL);
			if (pre_fd == -1)
				lerr(1, "invoke_m4():\n\t'%s' cannot open(): ",
				    filename, strerror(errno));

			/* replace stdin with the file prior to m4ing */
			if (dup2(pre_fd, STDIN_FILENO) == -1)
				lerr(1, "invoke_m4():\n\tcannot dup2(): %s",
				    strerror(errno));
		}

		/* and stdout is our temp file */
		if (dup2(post_fd, STDOUT_FILENO) == -1)
			lerr(1, "invoke_m4():\n\tcannot dup2() 2: %s",
			    strerror(errno));

		if (execlp("m4", "m4",
		    "-I" REL_M4_INCLUDE_PATH,
		    "-I.",
		    "-I" DATAROOTDIR "/l2cstd",
		    "-I" BUILD_DIR "/share/l2cstd", (char *) 0) == -1)
			lerr(1, "invoke_m4():\n\tcannot execlp(): %s",
			    strerror(errno));

	} else { /* parent */
		if (wait(&status) == -1 )
			lerr(1, "invoke_m4():\n\tcannot wait(): %s",
			    strerror(errno));

		if (status != 0)
			lerr(1, "invoke_m4():\n\tm4 returned non-zero: %s",
			    strerror(errno));
	}

	return new_filename;
}

/* cleans everything that was ever allocated */
void
free_everything()
{
	free_sat_results();
	free_sat_reports();
	free_cnf_clauses();
	free_literals();
	free_uq_elims();
	free_resolve_lits();
	free_nodes();
	free_maxmin_lits();

	if (yyin_filename != NULL) {
		free(yyin_filename);
		yyin_filename = NULL;
	}
}

/*
 * standard way of reporting errors and warnings
 */
void
lerr(int do_exit, const char *fmt, ...)
{
	va_list		ap;

	va_start(ap, fmt);
	if (do_exit)
		fprintf(stderr, "*error  : ");
	else 
		fprintf(stderr, "*warning: ");

	vfprintf(stderr, fmt, ap);
	fprintf(stderr, "\n");

	va_end(ap);
	if (do_exit) {
		free_everything();
		exit(EXIT_FAILURE);
	}
}

void
usage()
{
	fprintf(stderr, "+----------------------------------------------->\n");
	fprintf(stderr, "| Logic2CNF - MiniSAT on Steroids!\n");
	fprintf(stderr, "| Version: " PACKAGE_VERSION "\n");
	fprintf(stderr, "| Build: " __DATE__ " " __TIME__ "\n");
	fprintf(stderr, "+----------------------------------------------->\n");
	fprintf(stderr, "| (c) 2010 University of Kent\n");
	fprintf(stderr, "| Edd Barrett <e.barrett@kent.ac.uk>\n");
	fprintf(stderr, "+----------------------------------------------->\n");
	fprintf(stderr, "\n");
	fprintf(stderr, "Usage: logic2cnf [options] [file]\n");
	fprintf(stderr, "  -c		Just spit out converted DIMACS.\n");
	fprintf(stderr, "  -d		Enable debug info.\n");
	fprintf(stderr, "  -g		Dump graphviz for initial+converted tree.\n");
	fprintf(stderr, "  -h		Show this help.\n");
	fprintf(stderr, "  -j num	Show up to 'num' SAT results\n");
	fprintf(stderr, "  -l		Just spit out converted logic.\n");
	fprintf(stderr, "  -m		Enable M4 macro support.\n");
	fprintf(stderr, "  -p		Disable progress report.\n");
	fprintf(stderr, "  -r		Enable experimental optimiser.\n");
	fprintf(stderr, "  -s		Print stats only.\n");
	fprintf(stderr, "  -x		Leave behind m4 intermediate file.\n");
}

/* start here */
int
main(int argc, char **argv)
{
       	char			ch, *new_yyin_filename;
	lnode			*converted_nodes;
	int 			m4 = 0, no_unlink = 0, lim;
	int			mode = MODE_SOLVE, gv = 0;
	literal			*input;

	/* parse input from shell */
	while ((ch = getopt(argc, argv, "cdghj:lmprsx")) != -1) {
		switch (ch) {
			case 'c':
				mode = MODE_DIMACS;
				break;
			case 'd':
				l_debug = 1; /* turn on debug */
				break;
			case 'g':
				gv = 1;
				break;
			case 'j':
				lim = atol(optarg);
				if (lim < 1) {
					fprintf(stderr, "-j option incorrect\n");
					usage();
					exit(1);
				}
				sat_result_limit = (unsigned long int) lim;
				break;
			case 'l':
				mode = MODE_CNF;
				break;
			case 'm':
				m4 = 1;
				break;
			case 'p':
				enable_progress_report = 0;
				break;
			case 'h':
			default:
				usage();
				exit(1);
				break;
			case 'r':
				resolve_temps = 1;
				break;
			case 's':
				mode = MODE_STATS;
				break;
			case 'x':
				no_unlink = 1;
				break;
		}
	}

	argc -= optind;
	argv += optind;

	if (argc > 1) {
		lerr(0, "main():\n\ttrailing junk on command line");
		usage();
		exit(EXIT_FAILURE);
	}

	/* a filename was specified */
	if (argc == 1)
		yyin_filename = strdup(argv[0]);

	/* run m4 over the file and replace the temp file as the input file */
	if (m4) {
		new_yyin_filename = invoke_m4(yyin_filename);

		if (yyin_filename != NULL)
			free(yyin_filename);

		yyin_filename = new_yyin_filename;
	}

	/* if reading from a file, tell the parser which one */
	if (yyin_filename != NULL) {
		yyin = fopen(yyin_filename, "r");

		if (yyin == NULL)
			lerr(1, "main():\n\t'%s' cannot fopen(): %s",
			    yyin_filename,strerror(errno));
	} else
		yyin_filename = strdup("<stdin>");

	/* go away and parse some input. Express please! */
	SIMPLEQ_INIT(&sat_reports);
	SLIST_INIT(&uq_elims);
	SLIST_INIT(&resolve_lits);
	yyparse();
	my_yylineno = -1; /* lits defined here on are not in the source */

	/* clean up what we just parsed */
	fclose(yyin);
	if (m4) {
		if (!no_unlink) {
			if (unlink(yyin_filename) == -1)
				lerr(1, "main():\n\tcannot unlink() "
				    "m4 tmp file %s", strerror(errno));
		}
	}

	/* print some garbage if debug flag is set */
	if (l_debug) {
		fprintf(stderr, "\nInitially the tree looks like this:\n");
		dump_tree(last_node);
		fprintf(stderr, "\n");

	}

	/* make graphviz if -g */
	if (gv) {
		gv_init(GV_PT_FILENAME);
		gv_draw_tree(last_node);
		gv_done();
	}

	/* loop LITS checking all got used */
	RB_FOREACH(input, cnf_tree, &literals) {
		if (input->used == 0) {
			lerr(0, "%s line %d:\n\tunused literal '%s'",
			    yyin_filename, input->line_defined,
			    input->name);
			num_symbol_errors ++;
		}
	}

	/* if any lits were undefined or redefined stop */
	if (num_symbol_errors)
		lerr(1, "stopped due to %d literal symbol error(s)",
		    num_symbol_errors);

	/* traverse and convert */
	converted_nodes = start_cnf_conversion(last_node);

	/* print the converted tree for debug */
	if (l_debug) {
		fprintf(stderr, "After converting, tree is like this:\n");
		dump_tree(converted_nodes);
		fprintf(stderr, "\n");
	}

	/* make graphviz for converted tree if needed */
	if (gv) {
		gv_init(GV_AST_FILENAME);
		gv_draw_tree(converted_nodes);
		gv_done();
	}

	/* flatten the tree into a 2d table */
	TAILQ_INIT(&cnf_clauses);
	make_cnf_clauses(converted_nodes, 0);

	/* clean up before any optimisation etc. */
	if ((num_uq_elims) || (resolve_temps) || (num_resolve_lits))
		clean_cnf_clauses();

	/* resolve lits - needed for uq elimination also */
	if ((resolve_temps) || (num_resolve_lits))
		resolve();

	/* do any global quantifier elimination */
	if (num_uq_elims)
		remove_uq_lits_from_cnf_clauses();

	if ((num_uq_elims) || (resolve_temps) || (num_resolve_lits)) {
		/*
		 * superset removal only safe here!
		 * See comment in clean_cnf_clauses()
		 */
		mark_supersets_dead();
		if (enable_progress_report)
			fprintf(stderr, "Progress: cleaning CNF\n");

		clean_cnf_clauses(); /* uq can make duplicates */
		if (enable_progress_report)
			fprintf(stderr,
			    "Progress: Final optimisations phase yields "
			    "%lu clauses\n", num_cnf_clauses);
	}

	switch (mode) {
		case MODE_SOLVE:
			/* MiniSAT fun! */
			assign_minisat_lits();
			load_all_minisat_clauses();

			TAILQ_INIT(&sat_results);
			if (maxmin_mode == MAXMIN_UNDEFINED)
				solve_using_minisat();
			else {
				if (maxmin_mode == MAXMIN_MAX)
					solve_max_using_minisat();
				else
					solve_min_using_minisat();
			}
			break;
		case MODE_DIMACS:
			/* need to assign numbers even though not solving */
			assign_minisat_lits();
			dimacs_header();
			all_cnf_clauses_to_dimacs();
			break;
		case MODE_CNF:
			rename_internal_witness_lits();
			all_cnf_clauses_to_text(0);
			break;
		case MODE_STATS:
			print_stats();
			break;
		default:
			lerr(1, "main():\n\tunknown mode");
			break;
	};

	if (num_sat_reports)
		dump_all_sat_results_via_reports();
	else
		dump_all_raw_sat_results();

	/* Done, now be nice and free your memory */
	free_everything();

	return(EXIT_SUCCESS);
}

void
print_stats()
{
	printf("NPN:%35s: %lu\n", "number of parse nodes", num_nodes);
	printf("NL :%35s: %lu\n", "number of literals", num_literals);
	printf("NLU:%35s: %lu\n", "lits of which user",
	    num_literals - num_witness);
	printf("NLT:%35s: %lu\n", "lits of which tmp (for CNF conv)",
	    num_witness);
	printf("NCC:%35s: %lu\n", "number of CNF clauses", num_cnf_clauses);
}

/* Yay */
