/* ------------------
 * Logic2CNF
 * $Id: logic2cnf.h 509 2010-06-11 10:29:55Z eb771 $
 *
 * Copyright (c) 2009, Edd Barrett (University of Kent)
 * <eddbarrett at googlemail dot com>
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#ifndef LOGIC2CNF_H_
#define LOGIC2CNF_H_

#include <Solver.h>		/* !CXX! */
#include <SolverTypes.h>	/* !CXX! */

/* OpenBSD queues/trees - thanks guys */
#include "queue.h"
#include "tree.h"

#define M4_TMP_TEMPLATE		"tmp.logic2cnf.XXXXXXXX"
#define REL_M4_INCLUDE_PATH	"share/l2cstd/"

/* Tool modes of execution */
#define MODE_SOLVE	0
#define MODE_DIMACS	1
#define MODE_CNF	2
#define MODE_STATS	3

#define GV_PT_FILENAME		"pt.dot"
#define GV_AST_FILENAME		"ast.dot"

/* Assignment polarities */
#define POLARITY_NEG		0
#define POLARITY_POS		1
#define POLARITY_UNDEFINED	2

/* Maximal/minimal solving modes */
#define MAXMIN_UNDEFINED	0
#define MAXMIN_MIN		1
#define MAXMIN_MAX		2

/* ---[ TYPEDEFS ]-------------------------------------------------------- */

/* a cnf literal */
typedef struct literal_ {
	RB_ENTRY(literal_)		entries;
	uint8_t				internal_witness;
	int				line_defined;
	Lit				*minisat_lit;	/* !CXX minisat */
	unsigned long int		cnf_input_num;	/* lit num in DIMACS */
	uint8_t				used;
	char				*name;
} literal;
RB_PROTOTYPE(cnf_tree, literal_, entries, cnf_cmp);

/* A generic node */
typedef struct lnode_ {
	SIMPLEQ_ENTRY(lnode_)		entries;	/* for all_nodes */
	unsigned long int		id;	/* unique id */
	int				type;	/* lexer type */
	literal				*lit;	/* if a lit, its lit def */
	struct lnode_			*left;
	struct lnode_			*right;
} lnode;

/* a literal in a list of report literals */
typedef struct sat_report_lit_ {
	SIMPLEQ_ENTRY(sat_report_lit_)	entries;
	literal				*lit;
} sat_report_lit;

/* represents a list of nodes to report upon in some way */
typedef struct sat_report_ {
	SIMPLEQ_ENTRY(sat_report_)		entries;
	short int				type;
	SIMPLEQ_HEAD(srpl_head, sat_report_lit_)lits;
	char					*comment;
} sat_report;

/* a generic list of nodes (as a simpleq, so that they remain in order) */
typedef struct cnf_clause_lit_ {
	TAILQ_ENTRY(cnf_clause_lit_)	entries;
	literal				*lit;
	uint8_t				negated;
	/* used when deciding if a clause is identical */
	uint8_t				node_matched;
} cnf_clause_lit;

/* cnf_clauses, flattened CNF tail-queue list items */
typedef struct cnf_clause_ {
	TAILQ_ENTRY(cnf_clause_)		entries;
	unsigned long int			len;
	/* used in resolution */
	TAILQ_ENTRY(cnf_clause_)		res_polarity_entries;
	TAILQ_HEAD(nl_head, cnf_clause_lit_)	lits;
	uint8_t					dead; /* optimised away? */
	/* has this already been asubject of de-duplication? */
	uint8_t					deduped;
} cnf_clause;

typedef struct sat_result_ {
	TAILQ_ENTRY(sat_result_)		entries;
	SIMPLEQ_HEAD(srl_head,sat_result_lit_)	result;
} sat_result;

typedef struct sat_result_lit_ {
	SIMPLEQ_ENTRY(sat_result_lit_)		entries;
	literal					*lit;
	uint8_t					polarity;
} sat_result_lit;

/* list of literals to eliminate using global quantifier elimination */
typedef struct uq_elim_lit_ {
	SLIST_ENTRY(uq_elim_lit_)	entries;
	literal				*lit;
} uq_elim_lit;

/* list of literals to resolve at user's request */
typedef struct resolver_lit_ {
	SLIST_ENTRY(resolver_lit_)	entries;
	literal				*lit;
} resolver_lit;

/* for the list of max/min search literals */
typedef struct maxmin_lit_ {
	TAILQ_ENTRY(maxmin_lit_)	entries;
	uint8_t				polarity; /* once known */
	literal				*lit;
} maxmin_lit;

/* ---[ EXTERNS ]--------------------------------------------------------- */

/*
 * note, you can't extern list heads here. It just won't work.
 * instead, just extern on a per-file on demand basis.
 */

/* parser */
extern uint8_t			l_debug;
extern int			my_yylineno;
extern char			*yyin_filename;
extern FILE			*yyin;

/* nodes.c */
extern unsigned long int	num_nodes;
extern lnode			*last_node;

/* cnf_clauses.c */
extern unsigned long int 	num_witness;
extern unsigned long int 	num_literals;
extern unsigned long int	num_cnf_clauses;
extern unsigned long int	num_uq_elims;

/* resolve.c */
extern unsigned long int	num_resolve_lits;

/* logic2cnf.c */
extern int			num_symbol_errors;
extern uint8_t			l_debug;
extern uint8_t			resolve_temps;
extern uint8_t			enable_progress_report;
extern char			*yyin_filename;

/* sat.c */
extern unsigned long int	num_sat_results;
extern unsigned long int	sat_result_limit;
extern unsigned long int	num_sat_reports;
extern unsigned long int	num_maxmin_lits;
extern uint8_t			maxmin_mode;
extern uint8_t			num_maxmin_statements;

/* ---[ FUNCTIONS PROTOS ]------------------------------------------------ */

/* parser/lexer functions */
int 	yyparse();
int 	yylex();
void	yydbg(const char *);
void	yyerror(const char *);

/* logic2cnf.c */
void	dump_tree(lnode *node);
void	banner();
void	usage();
char	*invoke_m4();
void	free_everything();
void    lerr(int exit, const char *fmt, ...);
void	print_stats();

/* sat.c */
void	free_literals_lookup_table();
void	make_literal_lookup_table();
void	load_all_minisat_clauses();
void	load_minisat_clause(cnf_clause *clause);
void	solve_using_minisat();
void	add_counter_clause();
void	record_sat_result();
void	assign_minisat_lits();
void	free_sat_results();
void	free_sat_reports();
void	add_sat_report(short int type, char *comment);
void	add_lit_to_sat_report(char *nodename);
void	dump_all_raw_sat_results();
void	dump_raw_sat_result(sat_result *res);
void	dump_all_sat_results_via_reports();
void	dump_sat_result_via_report(sat_report rep, sat_result *res);
void	dump_sat_result_via_report_dec(sat_report *rep, sat_result *res);
void	add_maxmin_lit(char *litname, uint8_t mode);
void	free_maxmin_lits();
void	solve_min_using_minisat();
void	solve_max_using_minisat();
void	new_maxmin_statement();

/* nodes.c */
lnode	*new_node(int type, char *name, lnode *left, lnode *right);
lnode	*new_and_node(lnode *left, lnode *right);
lnode	*new_or_node(lnode *left, lnode *right);
lnode	*new_lit_node(char *name);
lnode	*new_witness_node();
lnode	*new_not_node(lnode *node);
lnode	*new_imply_node(lnode *witness, lnode *logic);		/* pseudo */
lnode	*new_bi_imply_node(lnode *witness, lnode *logic);	/* pseudo */
lnode	*new_xor_node(lnode *left, lnode *right);		/* pseudo */
void	free_nodes();
lnode	*get_node_subject(lnode *node, uint8_t *neg);
int	cnf_clause_len(cnf_clause *clause);
char	*get_node_str(lnode *node);
uint8_t	is_leaf(lnode *node);
void	expand_node_pool();

/* lits.c */
int		cnf_cmp(literal *c1, literal *c2);
void		free_literals();
literal	*define_lit(char *name, uint8_t internal);
literal	*get_literal(char *litname);
void		rename_internal_witness_lits();
literal	*get_literal_by_num(unsigned long int input);
void		free_literal(literal *lit);

/* to_cnf.c */
lnode	*to_cnf(lnode *node, lnode *witness);	/* wrapper for others */
lnode	*and_to_cnf(lnode *src, lnode *witness);
lnode	*or_to_cnf(lnode *src, lnode *witness);
lnode	*not_to_cnf(lnode *src, lnode *witness);
lnode	*lit_to_cnf(lnode *src, lnode *witness);
lnode 	*start_cnf_conversion(lnode *start);

/* cnf_clauses.c */
void	make_cnf_clauses(lnode *node, int *clause_ct);
void	add_lit_to_cur_cnf_clause(lnode *node);
void	insert_cur_cnf_clause();
void	free_cnf_clauses();
void	free_cnf_clause(cnf_clause *clause);

void		remove_dead_cnf_clauses();
uint8_t		cnf_clause_is_trivially_true(cnf_clause *clause);
void		mark_dead_trivially_true_cnf_clauses();
void		mark_identical_cnf_clauses_dead();
uint8_t 	cnf_clauses_are_identical(cnf_clause *clause1,
		    cnf_clause *clause2);
cnf_clause_lit	*find_lit_in_cnf_clause(cnf_clause_lit *lit,
		    cnf_clause *clause);
cnf_clause_lit	*find_lit_in_cnf_clause(literal *lit, cnf_clause *clause);
cnf_clause_lit	*find_lit_in_cnf_clauses(literal *lit);
void		remove_duplicate_literals_from_cnf_clause(cnf_clause *clause);
void		remove_duplicate_literals_from_cnf_clauses();
void		clean_cnf_clauses();
void		mark_dead_resolution_leftovers(literal *lit);

void		dimacs_header();
void		all_cnf_clauses_to_dimacs();
void		cnf_clause_to_dimacs(cnf_clause *clause);
void		all_cnf_clauses_to_text(uint8_t to_stderr);
void		cnf_clause_to_text(cnf_clause *clause, uint8_t to_stderr);

/* resolve.c */
void		resolve();
void		add_resolve_lit(char *litname);
void		resolve_user_lits();
void		resolve_temp_lits();
void		try_resolve_clauses(cnf_clause *pri,
		    cnf_clause *sec, literal *lit);
void		resolve_combine_and_mark_dead(cnf_clause *pri,
		    cnf_clause *sec, cnf_clause_lit *pri_item,
		    cnf_clause_lit *sec_item);
void		resolve_lit(literal *lit);
void		free_resolve_lits();
void		mark_supersets_dead();
cnf_clause	*one_clause_is_superset(cnf_clause *pri, cnf_clause *sec);
void		construct_polarity_lists(literal *lit);

/* uq_elims.c */
void	free_uq_elims();
void	add_uq_elim(char *lit_name);
void	remove_uq_lits_from_cnf_clauses();
void	remove_uq_lits_from_clause(uq_elim_lit *lit, cnf_clause *clause);

/* graphviz.c */
void			gv_init(const char *filename);
void			gv_done();
unsigned long int	gv_draw_tree(lnode *node);

/* util.c */
void	lerr(int do_exit, const char *fmt, ...);
void    *xmalloc(size_t sz);
void    *xrealloc(void *ptr, size_t sz);
#ifndef HAVE_ASPRINTF
int     asprintf(char **ret, const char *fmt, ...);
#endif

#endif
