/* ------------------
 * Logic2CNF
 * $Id: nodes.c 515 2010-06-11 11:15:11Z eb771 $
 *
 * Copyright (c) 2010, Edd Barrett (University of Kent)
 * <e.barrett@kent.ac.uk>
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include "logic2cnf.h"
#include "y.tab.h"
#include "queue.h"

/* pool of nodes - allocated all at once = quicker */
#define INIT_NODE_POOL_SZ	1024
unsigned long int		node_pool_sz = 0;
unsigned long int		node_pool_used = 0;
lnode				*node_pool = NULL;

typedef struct node_pool_seg_ {
	SIMPLEQ_ENTRY(node_pool_seg_)	entries;
	lnode				*seg;
} node_pool_seg;
unsigned long int			num_node_pool_segs = 0;
SIMPLEQ_HEAD(nphead,node_pool_seg_)	node_pool_segs_head;

lnode					*last_node = NULL;
unsigned long int 			num_nodes = 0;

void
expand_node_pool()
{
	node_pool_seg		*seg;

	if (node_pool_sz == 0)	{ /* unused as of yet */
		node_pool_sz = INIT_NODE_POOL_SZ;
		SIMPLEQ_INIT(&node_pool_segs_head);
	} else {
		node_pool_sz *= 2;
	}

	node_pool_used = 0;
	
	seg = (node_pool_seg *) xmalloc(sizeof(node_pool_seg));
	node_pool = (lnode *) xmalloc(node_pool_sz * sizeof(lnode));
	seg->seg = node_pool;

	SIMPLEQ_INSERT_HEAD(&node_pool_segs_head, seg, entries);
}

/* get a new node to add to the tree */
lnode *
new_node(int type, char *name, lnode *left, lnode *right)
{
       	literal			*literal = NULL;
	char			*witness_litname;
	lnode			*node;

	/* check we are using this correctly */
	if (type == NOT)
		assert(right == NULL);
	else if ((type == AND) || (type == OR))
		assert((left != NULL) && (right != NULL));

	if (type == LIT) {
		/* see if this lit is defined */
		literal = get_literal(name);

		if (literal == NULL) { /* if this literal is not yet defd */

			/* all non-witness lits should be defined by now */
			if (name != NULL) {
				lerr(0, "%s line %d:\n"
				    "\tliteral '%s' undefined",
				    yyin_filename, my_yylineno, name);
				num_symbol_errors ++;

				/*
				 * We define it anyway, so that it can still
				 * be used in the parse tree atleast. A check
				 * in main will catch this later and exit.
				 */
				literal = define_lit(name, 0);
			} else {
				/* witness literal needs to be defined */
				if (asprintf(&witness_litname,
				    "_x%lu", num_witness ++) == -1) {
					lerr(1, "new_node():\n\tcannot "
				        "asprintf(): %s", strerror(errno));
				}
				literal = define_lit(witness_litname, 1);
				free(witness_litname);
			}
		}

		literal->used = 1; /* mark used */

	} /* else literal remains NULL */

	/* yoink a node from the pool */
	if (node_pool_used == node_pool_sz)
		expand_node_pool();

	node = &node_pool[node_pool_used ++];

	/* as tempting as it is to zero this, it gets slow on loads of nodes */
	/* memset(node, 0, sizeof(lnode)); */
	node->type = type;
	node->id = ++ num_nodes;
	node->left = left;
	node->right = right; /* not used for NOT */
	node->lit = literal;

	last_node = node;
	return node;
}

/* create a new binary AND node */
lnode *
new_and_node(lnode *left, lnode *right)
{
	return new_node(AND, NULL, left, right);
}

/* create a new binary OR node */
lnode *
new_or_node(lnode *left, lnode *right)
{
	return new_node(OR, NULL, left, right);
}

/* create a new literal node */
lnode *
new_lit_node(char *name)
{
	return new_node(LIT, name, NULL, NULL);
}

/* create a new literal node as a witness */
lnode *
new_witness_node()
{
	return new_node(LIT, NULL, NULL, NULL);
}

/* create a new unary NOT node */
lnode *
new_not_node(lnode *node)
{
	return new_node(NOT, NULL, node, NULL);
}

/* create a implication node (right =>, swap params for left) */
lnode *
new_imply_node(lnode *witness, lnode *logic)
{
	return new_or_node(new_not_node(witness), logic);
}

/* create a bi-implication node */
lnode *
new_bi_imply_node(lnode *witness, lnode *logic)
{
	lnode	*limply, *rimply;

	rimply = new_imply_node(witness, logic);
	limply = new_imply_node(logic, witness);

	return new_and_node(limply, rimply);
}


/* create a pseudo XOR node */
lnode *
new_xor_node(lnode *left, lnode *right)
{
	lnode	*clause1, *clause2;

	clause1 = new_and_node(left, new_not_node(right));
	clause2 = new_and_node(new_not_node(left), right);
	return new_or_node(clause1, clause2);
}


int
cnf_clause_len(cnf_clause *clause)
{
	cnf_clause_lit		*item;
	int			ct = 0;

	TAILQ_FOREACH(item, &clause->lits, entries)
		ct ++;
	return ct;
}

/*
 * get the actual lit node and strip any negations. If negated the ptr
 * 'neg' is set on return. Try not to use this much. Use cached
 * fields in cnf_clause struct instead.
 */
lnode *
get_node_subject(lnode *node, uint8_t *neg)
{
	lnode			*sj = node;

	*neg = 0;
	if (node->type == NOT) {
		*neg = 1;
		sj = node->left;
	}

	return sj;
}

uint8_t
is_leaf(lnode *node)
{
	if (node->type == LIT)
		return 1;

	return 0;
}

/* makes a user readable description of a node, must free() result! */
char *
get_node_str(lnode *node)
{
	char	*ret = NULL;
	int	err = 0;

	/* XXX dont use asprintf! */
	switch (node->type) {
	case NOT:
		err = asprintf(&ret, "~");
		break;
	case AND:
		err = asprintf(&ret, ".");
		break;
	case OR:
		err = asprintf(&ret, "+");
		break;
	case LIMPL:
		err = asprintf(&ret, "<=");
		break;
	case RIMPL:
		err = asprintf(&ret, "=>");
		break;
	case BIMPL:
		err = asprintf(&ret, "<=>");
		break;
	case LIT:
		err = asprintf(&ret, "%s", node->lit->name);
		break;
	default:
		lerr(1, "get_node_str():\n\tunknown node type: %d", node->type);
	};

	if (err == -1)
		lerr(1, "get_node_str():\n\tasprintf()");

	return ret;
}

void free_nodes()
{
	node_pool_seg 		*seg, *next_seg;

	for (seg = SIMPLEQ_FIRST(&node_pool_segs_head);
	    seg != SIMPLEQ_END(&node_pool_segs_head);
	    seg = next_seg) {
		next_seg = SIMPLEQ_NEXT(seg, entries);
		free(seg);	
	}
}
