/* ------------------
 * Logic2CNF
 * $Id: resolve.c 515 2010-06-11 11:15:11Z eb771 $
 *
 * Copyright (c) 2010, Edd Barrett (University of Kent)
 * <e.barrett@kent.ac.uk>
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "logic2cnf.h"
#include "y.tab.h"
#include "queue.h"

/* from lits.c */
extern RB_HEAD(cnf_tree, literal_)		literals;

/* from cnf_clauses.c */
extern TAILQ_HEAD(ncs_head, cnf_clause_)	cnf_clauses;

SLIST_HEAD(rsl_head, resolver_lit_)		resolve_lits;
unsigned long int				num_resolve_lits = 0;

/*
 * list containing +ve and -ve instances of the literal we
 * are about to resolve
 */
TAILQ_HEAD(pos_ncs_head, cnf_clause_)		pos_clauses;
TAILQ_HEAD(neg_ncs_head, cnf_clause_)		neg_clauses;

void
construct_polarity_lists(literal *lit)
{
	cnf_clause_lit		*cl_lit;
	cnf_clause		*clause;

	TAILQ_INIT(&pos_clauses);
	TAILQ_INIT(&neg_clauses);

	TAILQ_FOREACH(clause, &cnf_clauses, entries) {
		cl_lit = find_lit_in_cnf_clause(lit, clause);

		if (cl_lit == NULL)
			continue;

		if (cl_lit->negated) {
			TAILQ_INSERT_TAIL(&neg_clauses, clause,
			    res_polarity_entries);
		} else {
			TAILQ_INSERT_TAIL(&pos_clauses, clause,
			    res_polarity_entries);
		}
	}
}

void
resolve()
{
	if (resolve_temps)
		resolve_temp_lits();

	if (num_resolve_lits)
		resolve_user_lits();
}

/* add a user-elected literal to be resolved */
void
add_resolve_lit(char *litname)
{
	literal			*lit;
	resolver_lit		*rl;

	lit = get_literal(litname);

	if (lit == NULL) {
		lerr(0, "%s line %d:\n\tliteral '%s' undefined in resolve "
		    " statement.\n", yyin_filename, my_yylineno, litname);
		num_symbol_errors ++;
		return;
	}

	rl = (resolver_lit *) xmalloc(sizeof(resolver_lit));
	rl->lit = lit;

	SLIST_INSERT_HEAD(&resolve_lits, rl, entries);
	num_resolve_lits ++;

}

/* resolve literals that the user requested to resolve */
void
resolve_user_lits()
{
	resolver_lit		*rl;
	unsigned long int	done = 1, prev_num_clauses = num_cnf_clauses;

	SLIST_FOREACH(rl, &resolve_lits, entries) {

		if (enable_progress_report) {
			fprintf(stderr, "\rProgress: Resolve user lit "
			    "%8lu/%8lu (%8lu -> %8lu clauses)",
			    done, num_resolve_lits,
			    prev_num_clauses, num_cnf_clauses);
			if (l_debug) {
				fprintf(stderr, "\nAbout to resolve %s\n",
				    rl->lit->name);
				fprintf(stderr, "CNF thus far:\n");
				all_cnf_clauses_to_text(1);
			}
		}

		/* get lists ready */
		construct_polarity_lists(rl->lit);

		resolve_lit(rl->lit);
		mark_dead_resolution_leftovers(rl->lit);
		clean_cnf_clauses();
		done ++;

		/* literal should be dead now */
		RB_REMOVE(cnf_tree, &literals, rl->lit);
		free_literal(rl->lit);
		num_literals --;
	}

	if (enable_progress_report)
		fprintf(stderr, "\n");
}

/*
 * use propositional logic resolution on temp lits in order to make a
 * far more compressed CNF representation which is equivilent.
 */
void
resolve_temp_lits()
{
	literal				*item, *next_item;
	unsigned long int		done = 1;
	unsigned long int		prev_num_clauses = num_cnf_clauses;


	for (item = RB_MIN(cnf_tree, &literals);
	    item != NULL;
	    item = next_item) {

	    next_item = RB_NEXT(cnf_tree, &literals, item);
		/* for temp literals starting _x only */
		if (! item->internal_witness)
			continue;

		if (enable_progress_report) {
			fprintf(stderr, "\rProgress: Resolve temp lit "
			    "%8lu/%8lu (%8lu -> %8lu clauses)",
			    done, num_witness,
			    prev_num_clauses, num_cnf_clauses);
			if (l_debug) {
				fprintf(stderr, "\nAbout to resolve %s\n",
				    item->name);
				fprintf(stderr, "CNF thus far:\n");
				all_cnf_clauses_to_text(1);
			}
		}

		/* get queues ready */
		construct_polarity_lists(item);

		/* do the resolution dance on each literals */
		resolve_lit(item);
		mark_dead_resolution_leftovers(item);
		clean_cnf_clauses();

		/* in theory literals has gone now */
		RB_REMOVE(cnf_tree, &literals, item);
		free_literal(item);
		num_literals --;

		done ++;
	}

	if (enable_progress_report)
		fprintf(stderr, "\n");
}

/*
 * resolve a single literal in cnf_clauses.
 */
void
resolve_lit(literal *lit)
{
	cnf_clause		*primary, *secondary;

	TAILQ_FOREACH(primary, &pos_clauses, res_polarity_entries) {
		TAILQ_FOREACH(secondary, &neg_clauses, res_polarity_entries) {
			try_resolve_clauses(primary, secondary, lit);
		}
	}
}


/* will these two clauses resolve for the literal *lit? */
void
try_resolve_clauses(cnf_clause *pri, cnf_clause *sec,
    literal *lit)
{
	cnf_clause_lit			*pri_item, *sec_item;

	TAILQ_FOREACH(pri_item, &pri->lits, entries) {

		/* if not the right lit, we dont care */
		if (pri_item->lit != lit)
			continue;

		TAILQ_FOREACH(sec_item, &sec->lits, entries) {

			/* can't resolve yourself */
			if (pri_item == sec_item)
				continue;

			if (sec_item->lit != lit)
				continue;

			/* at this stage we have matching literals atleast */
			if (pri_item->negated != sec_item->negated) {
				/* if complementary, can! */
				resolve_combine_and_mark_dead(pri, sec,
				    pri_item, sec_item);
				break;
			}
		}
	}
}

/*
 * resolve two clauses. Remove complimented lits and combine to make new 
 * clause. The old ones are marked dead and removed later.
 */
void
resolve_combine_and_mark_dead(cnf_clause *pri, cnf_clause *sec,
    cnf_clause_lit *pri_item, cnf_clause_lit *sec_item)
{
	cnf_clause_lit		*item, *new_item;
	cnf_clause		*new_clause;
	unsigned long int	new_len = 0;

	if (l_debug) {
		fprintf(stderr, "resolving '%s' against clauses:\n",
		    pri_item->lit->name);
		fprintf(stderr, "  1) ");
		cnf_clause_to_text(pri, 1);
		fprintf(stderr, "\n  2) ");
		cnf_clause_to_text(sec, 1);
		fprintf(stderr, "\n");
	}

	/* make space for a linked list of new nodes for this clause */
	new_clause = (cnf_clause *) xmalloc(sizeof(cnf_clause));
	TAILQ_INIT(&new_clause->lits);
	new_clause->dead = 0;
	new_clause->deduped = 0;

	/*
	 * the new clause is all lits in the old clause apart from the
	 * two that allowed resolution, in this case, pri_item, sec_item
	 */
	TAILQ_FOREACH(item, &pri->lits, entries) {
		if (item == pri_item)
			continue;

		new_item = (cnf_clause_lit *) xmalloc(sizeof(cnf_clause_lit));

		new_item->negated = item->negated;
		new_item->lit = item->lit;

		TAILQ_INSERT_TAIL(&new_clause->lits, new_item, entries);
		new_len ++;
	}

	TAILQ_FOREACH(item, &sec->lits, entries) {
		if (item == sec_item)
			continue;

		new_item = (cnf_clause_lit *) xmalloc(sizeof(cnf_clause_lit));

		new_item->negated = item->negated;
		new_item->lit = item->lit;

		TAILQ_INSERT_TAIL(&new_clause->lits, new_item, entries);
		new_len ++;
	}

	/* assign new length */
	new_clause->len = new_len;

	/* we may well have duplicates here */
	remove_duplicate_literals_from_cnf_clause(new_clause);

	if (l_debug) {
		fprintf(stderr, "=> ");
		cnf_clause_to_text(new_clause, 1);
		fprintf(stderr, "\n");
	}

	/* important! remove this step and memeory goes BANG */
	if ((new_clause->len == 0) ||
	    (cnf_clause_is_trivially_true(new_clause))) {
		free_cnf_clause(new_clause);
	} else {
		TAILQ_INSERT_TAIL(&cnf_clauses, new_clause, entries);
		num_cnf_clauses++;
	}

	/* mark dead ones for deletion */
	pri->dead = 1;
	sec->dead = 1;
}

void
free_resolve_lits()
{
	resolver_lit *r, *next;

	for (r = SLIST_FIRST(&resolve_lits);
	    r != SLIST_END(&resolver_lits);
	    r = next) {

		next = SLIST_NEXT(r, entries);
		free(r);
	}
}

cnf_clause *
one_clause_is_superset(cnf_clause *pri, cnf_clause *sec)
{
	unsigned long int	plen, slen;
	cnf_clause		*longer, *shorter;
	cnf_clause_lit		*slit;

	plen = pri->len;
	slen = sec->len;

	if (plen == slen)
		return NULL;

	/* find the longest clause */
	if (plen > slen) {
		longer = pri;
		shorter = sec;
	} else {
		longer = sec;
		shorter = pri;
	}

	/* iterate over shorter of two and se if all lits appear in longer */
	TAILQ_FOREACH(slit, &shorter->lits, entries) {
		if (!find_lit_in_cnf_clause(slit, longer))
			return NULL;
	}

	return longer;
}


void
mark_supersets_dead()
{
	cnf_clause		*pri, *sec, *super;

	TAILQ_FOREACH(pri, &cnf_clauses, entries) {
		if (pri->dead)
			continue;

		TAILQ_FOREACH(sec, &cnf_clauses, entries) {

			if (sec->dead)
				continue;
			if (pri == sec)
				continue;

			super = one_clause_is_superset(pri, sec);
			if (super == NULL)
				continue;

			super->dead = 1;
		}
	}
}

/*
 * remove any clauses contaning the lit
 */
void
mark_dead_resolution_leftovers(literal *lit)
{
	cnf_clause		*clause, *next_clause;

	for (clause = TAILQ_FIRST(&cnf_clauses);
	    clause != TAILQ_END(&cnf_clauses);
	    clause = next_clause) {
		next_clause = TAILQ_NEXT(clause, entries);

		if (find_lit_in_cnf_clause(lit, clause)) {
			clause->dead = 1;
			if (l_debug) {
				fprintf(stderr, "removing leftover "
				    "resolution junk for '%s'\n", lit->name);
				cnf_clause_to_text(clause, 1);
				fprintf(stderr, "\n");
			}

		}
	}
}


