/* ------------------
 * Logic2CNF
 * $Id: sat.c 515 2010-06-11 11:15:11Z eb771 $
 *
 * Copyright (c) 2010, Edd Barrett (University of Kent)
 * <e.barrett@kent.ac.uk>
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <math.h>

#include "logic2cnf.h"
#include "y.tab.h"
#include "queue.h"

/* from cnf_clauses.c */
extern TAILQ_HEAD(ncs_head, cnf_clause_)	cnf_clauses;

/* from lits.c */
extern RB_HEAD(cnf_tree, literal_)		literals;

Solver					solver;	/* !CXX */
unsigned long int			sat_result_limit = 0;
TAILQ_HEAD(sr_head, sat_result_)	sat_results;
unsigned long int			num_sat_results = 0;

unsigned long int			num_sat_reports = 0;
sat_report				*cur_sat_report = NULL;
SIMPLEQ_HEAD(srp_head, sat_report_)	sat_reports;

/* once the literals TAILQ wont change unroll into array for fast lookup */
literal					**final_inputs = NULL;

/* maximum/miniumum solving support */
TAILQ_HEAD(mm_head, maxmin_lit_)	maxmin_lits;
unsigned long int			num_maxmin_lits = 0;
uint8_t					maxmin_mode = MAXMIN_UNDEFINED;
uint8_t					num_maxmin_statements = 0;

void
new_maxmin_statement()
{
	if (num_maxmin_statements > 0)
		lerr(1, "%s line %d:\n\tmore than one find_max/min used. abort.",
		    yyin_filename, my_yylineno);

	num_maxmin_statements ++;
}

/* take cnf_clauses and load into minisat */
void
load_all_minisat_clauses()
{
	cnf_clause	*clause;

	TAILQ_FOREACH(clause, &cnf_clauses, entries)
		load_minisat_clause(clause);
}

void
make_literal_lookup_table()
{
	literal		*inp;

	/* array is one bigger as 0 is an impossible cnf input number */
	final_inputs = (literal **)
	    xmalloc(sizeof(literal *) * (num_literals + 1));

	RB_FOREACH(inp, cnf_tree, &literals)
		final_inputs[inp->cnf_input_num] = inp;
}

void
free_literals_lookup_table()
{
	free(final_inputs);
}

/*
 * load a single clause into minisat
 * !CXX!
 */
void
load_minisat_clause(cnf_clause *clause)
{
	vec<Lit>		clause_m;
	cnf_clause_lit		*item;

	TAILQ_FOREACH(item, &clause->lits, entries) {
		if (!item->negated)
			clause_m.push(*(item->lit->minisat_lit));
		else
			clause_m.push(~(*(item->lit->minisat_lit)));
	}
	solver.addClause(clause_m);
}

/*
 * negate the last found result and add to solver
 * !CXX!
 */
void
add_counter_clause()
{
	int 		i;	/* minisat doesnt use long */
	Lit		*lit;
	vec<Lit>	clause;

	/* check lookup table is present */
	assert(final_inputs != NULL);

	for (i = 0; i < solver.model.size(); i ++) {
		lit = final_inputs[i + 1]->minisat_lit;

		if (solver.model[i] == l_True)
			clause.push(~(*lit));
		else
			clause.push(*lit);
	}

	solver.addClause(clause);
}

/*
 * yoink the result from the solver model and
 * translate back to lnodes.
 */
void
record_sat_result()
{
	sat_result			*item;
	sat_result_lit			*rlit;
	int				i;
	literal				*lit;

	/* check lookup table is present */
	assert(final_inputs != NULL);

	item = (sat_result *) xmalloc(sizeof(sat_result));
	SIMPLEQ_INIT(&item->result); /* list of literal assignments here */

	/* we dont care to store witness lits */
	for (i = 0; i < solver.model.size(); i ++) {
		lit = final_inputs[i + 1];

		rlit = (sat_result_lit *)
		    xmalloc(sizeof(sat_result_lit));

		rlit->lit = lit;
		if (solver.model[i] == l_True)
			rlit->polarity = 1;
		else
			rlit->polarity = 0;

		SIMPLEQ_INSERT_TAIL(&item->result, rlit, entries);
	}

	TAILQ_INSERT_TAIL(&sat_results, item, entries);	
	num_sat_results ++;
}

void
assign_minisat_lits()
{
	literal				*item;
	unsigned long int		i = 0;

	RB_FOREACH(item, cnf_tree, &literals) {
		item->cnf_input_num = ++ i;
		item->minisat_lit = new Lit(solver.newVar());
	}
}

/*
 * actually run MiniSAT
 * !CXX!
 */
void
solve_using_minisat()
{
	bool			sat = true;

	/* make lookup table for inputs - quick! */
	make_literal_lookup_table();

	solver.simplify();
	do {
		sat = solver.solve();
		if (sat) {
			add_counter_clause();
			record_sat_result();

			if (enable_progress_report)
				fprintf(stderr, "\rProgress: Solving soltion %lu",
				    num_sat_results);
			if (l_debug)
				fprintf(stderr, "\n");
		}

		if ((sat_result_limit != 0 ) &&
		    (num_sat_results >= sat_result_limit))
			break; /* reached result limit */

	} while (sat == true);

	free_literals_lookup_table();

	if (enable_progress_report)
		fprintf(stderr, "\n");
}

/*
 * put the cur_sat_report into the table and start a new one
 */
void
add_sat_report(short int type, char *comment)
{
	assert((type == REPORT_LITERALS) || (type == REPORT_DEC));

	/*
	 * if there were no lits in the report, then no sat_report
	 * would have been allocated, so we must make one.
	 */
	if (cur_sat_report == NULL) {
		cur_sat_report = (sat_report *) xmalloc(sizeof(sat_report));
		SIMPLEQ_INIT(&cur_sat_report->lits);
	}

	cur_sat_report->type = type;
	cur_sat_report->comment = NULL;

	if (comment !=  NULL) {
		/* is strdup()d in the lexer */
		cur_sat_report->comment = strdup(comment);
	}

	SIMPLEQ_INSERT_TAIL(&sat_reports, cur_sat_report, entries);
	num_sat_reports ++;
	cur_sat_report = NULL;	/* ready for new alloc */
}

void
add_lit_to_sat_report(char *nodename)
{
	literal				*lit;
	sat_report_lit			*sr_lit;

	lit = get_literal(nodename);
	if (lit == NULL) {
		lerr(0, "%s line %d:\n\tundefined literal '%s' in report",
		    yyin_filename, my_yylineno, nodename);
		num_symbol_errors ++;
	}

	/*
	 * if this is the first literal for this report,
	 * we must allocate space for the report
	 */
	if (cur_sat_report == NULL) {
		cur_sat_report = (sat_report *) xmalloc(sizeof(sat_report));
		SIMPLEQ_INIT(&cur_sat_report->lits);
	}

	sr_lit = (sat_report_lit *) xmalloc(sizeof(sat_report_lit));
	sr_lit->lit = lit;

	SIMPLEQ_INSERT_TAIL(&cur_sat_report->lits, sr_lit, entries);
}

/*
 * dump a single result and it's witness lits
 */
void
dump_raw_sat_result(sat_result *res)
{
	sat_result_lit		*sr_lit;
	
	SIMPLEQ_FOREACH(sr_lit, &res->result, entries) {
		if (sr_lit->lit->internal_witness)
			continue;

		if (sr_lit->polarity)
			printf(" %s ", sr_lit->lit->name);
		else
			printf("~%s ", sr_lit->lit->name);
	}
}

/*
 * dump raw sat results output excluding witness literals
 */
void
dump_all_raw_sat_results()
{
	sat_result		*item;
	unsigned long int	i = 0;

	TAILQ_FOREACH(item, &sat_results, entries) {
		printf("%lu: ", i++);
		dump_raw_sat_result(item);	
		printf("\n");
	}
}

void
dump_sat_result_via_report(sat_report *rep, sat_result *res)
{	
	sat_report_lit		*srp_lit;
	sat_result_lit		*res_lit;			

	SIMPLEQ_FOREACH(srp_lit, &rep->lits, entries) {
		SIMPLEQ_FOREACH(res_lit, &res->result, entries) {

			if (srp_lit->lit != res_lit->lit)
				continue;

			if (res_lit->polarity)
				printf(" %s ", res_lit->lit->name);
			else
				printf("~%s ", res_lit->lit->name);
		}
	}
}

/*
 * total up a set of given sat inputs as if they represented an
 * unsigned decimal number. 
 */
void
dump_sat_result_via_report_dec(sat_report *rep, sat_result *res)
{
	unsigned long int		bit_val = 1, tot = 0;
	sat_report_lit			*srp_lit;
	sat_result_lit			*res_lit;

	SIMPLEQ_FOREACH(srp_lit, &rep->lits, entries) {
		SIMPLEQ_FOREACH(res_lit, &res->result, entries) {

			if (res_lit->lit != srp_lit->lit)
				continue;

			if (res_lit->polarity)
				tot += bit_val;

			bit_val *= 2;
		}
	}

	printf("%lu", tot);
}

void
dump_all_sat_results_via_reports()
{
	assert(num_sat_reports);

	sat_result			*result;
	sat_report			*report;
	unsigned long int		result_num = 0;

	TAILQ_FOREACH(result, &sat_results, entries) {
		SIMPLEQ_FOREACH(report, &sat_reports, entries) {

			if (report->comment == NULL) 
				printf("%lu: ", result_num);
			else
				printf("%lu: %s: ",
				    result_num, report->comment);

			if (report->type == REPORT_LITERALS)
				dump_sat_result_via_report(report, result);
			else
				dump_sat_result_via_report_dec(report, result);
			
			printf("\n");
		}
		result_num ++;

		if (result != TAILQ_LAST(&sat_results, sr_head))
			printf("\n");
	}
}

void
free_sat_results()
{

	sat_result		*sres_item, *next_sres;
	sat_result_lit		*sres_lit_item, *next_sres_lit_item;

	for (sres_item = TAILQ_FIRST(&sat_results);
	    sres_item != TAILQ_END(&sat_results);
	    sres_item = next_sres) {

		next_sres = TAILQ_NEXT(sres_item, entries);

		for (sres_lit_item = SIMPLEQ_FIRST(&sres_item->result);
		    sres_lit_item != SIMPLEQ_END(&sres_item->result);
		    sres_lit_item = next_sres_lit_item) {

			next_sres_lit_item =
			    SIMPLEQ_NEXT(sres_lit_item, entries);
			free(sres_lit_item);
		}
		free(sres_item);
	}
	num_sat_results = 0;
}

void
free_sat_reports()
{
	sat_report		*sat_rep, *next_sat_rep;
	sat_report_lit		*rep_lit, *next_rep_lit;

	for (sat_rep = SIMPLEQ_FIRST(&sat_reports);
	    sat_rep != SIMPLEQ_END(&sat_reports);
	    sat_rep = next_sat_rep) {
		next_sat_rep = SIMPLEQ_NEXT(sat_rep, entries);

		for (rep_lit = SIMPLEQ_FIRST(&sat_rep->lits);
		    rep_lit != SIMPLEQ_END(&sat_rep->lits);
		    rep_lit = next_rep_lit) {
			next_rep_lit = SIMPLEQ_NEXT(rep_lit, entries);
			free(rep_lit);
		}

		if (sat_rep->comment != NULL)
			free(sat_rep->comment);
		free(sat_rep);
	}
	num_sat_reports = 0;
}

/*
 * register literals for max/min solving
 */
void
add_maxmin_lit(char *litname, uint8_t mode)
{
	maxmin_lit		*mm_lit;
	literal			*input;

	/* don't mix find_max/find_min */
	if ((maxmin_mode != MAXMIN_UNDEFINED) &&
	    (maxmin_mode != mode))
		lerr(1, "%s line %d:\n\tfind_max/min may not be used "
		    "in conjunction.", yyin_filename, my_yylineno);

	/* if first, then init list also */
	if (maxmin_mode == MAXMIN_UNDEFINED)
		TAILQ_INIT(&maxmin_lits);

	maxmin_mode = mode;

	/* check defined */
	input = get_literal(litname);
	if (input == NULL) {
		lerr(0, "%s line %d:\n\tundefined literal '%s' in find_xxx.",
		    yyin_filename, my_yylineno, litname);
		num_symbol_errors ++;
	}

	/* go for it */
	mm_lit = (maxmin_lit *) xmalloc(sizeof(maxmin_lit));
	mm_lit->lit = input;
	mm_lit->polarity = POLARITY_UNDEFINED;

	/*
	 * msb should come first in this list, even though in l2c
	 * language you specify lsb first. Saves us reverse traversing
	 * later
	 */
	TAILQ_INSERT_HEAD(&maxmin_lits, mm_lit, entries);
	num_maxmin_lits ++;
}

void
free_maxmin_lits()
{
	maxmin_lit		*lit, *next;

	for (lit = TAILQ_FIRST(&maxmin_lits);
	    lit != TAILQ_END(&maxmin_lits);
	    lit = next) {
		next = TAILQ_NEXT(lit, entries);
		TAILQ_REMOVE(&maxmin_lits, lit, entries);
		free(lit);
	}

}

/*
 * find a maximum or minimum sat model
 * !CXX
 */
void
solve_min_using_minisat()
{
	bool			sat = true;
	maxmin_lit		*mm_lit;
	vec<Lit>		assumption;

	/* make lookup table for literals - quick! */
	make_literal_lookup_table();

	/* check the constraint is satisfiable in the first place */
	solver.simplify();
	if (solver.solve() != true)
		return;

	solver.simplify();
	TAILQ_FOREACH(mm_lit, &maxmin_lits, entries) {

		assumption.push(~*mm_lit->lit->minisat_lit);
		sat = solver.solve(assumption);

		/* if not SAT - negate the assumption */
		if (!sat) {
			assumption.clear();
			assumption.push(*mm_lit->lit->minisat_lit);
		}

		solver.addClause(assumption);
		solver.simplify();
		assumption.clear();
	}

	/* check the result is correct also */
	solver.simplify();
	if (solver.solve() != true)
		return;

	record_sat_result();
	free_literals_lookup_table();
}

void
solve_max_using_minisat()
{
	bool			sat = true;
	maxmin_lit		*mm_lit;
	vec<Lit>		assumption;

	/* make lookup table for literals - quick! */
	make_literal_lookup_table();

	/* check the constraint is satisfiable in the first place */
	solver.simplify();
	if (solver.solve() != true)
		return;

	solver.simplify();
	TAILQ_FOREACH(mm_lit, &maxmin_lits, entries) {

		assumption.push(*mm_lit->lit->minisat_lit);
		sat = solver.solve(assumption);

		/* if not SAT - negate the assumption */
		if (!sat) {
			assumption.clear();
			assumption.push(~*mm_lit->lit->minisat_lit);
		}

		solver.addClause(assumption);
		solver.simplify();
		assumption.clear();
	}

	/* check the result is correct also */
	solver.simplify();
	if (solver.solve() != true)
		return;

	record_sat_result();
	free_literals_lookup_table();
}
