/* ------------------
 * Logic2CNF
 * $Id: to_cnf.c 515 2010-06-11 11:15:11Z eb771 $
 *
 * Copyright (c) 2010, Edd Barrett (University of Kent)
 * <e.barrett@kent.ac.uk>
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "logic2cnf.h"
#include "y.tab.h"
#include "queue.h"

/* convert ANY node to CNF */
lnode *
to_cnf(lnode *node, lnode *witness)
{
	lnode		*new_node = NULL;

	/* i resisted the urge to use func pointers in the node struct */
	switch(node->type)  {
		case AND:
			new_node = and_to_cnf(node, witness);
			break;
		case OR:
			new_node = or_to_cnf(node, witness);
			break;
		case NOT:
			new_node = not_to_cnf(node, witness);
			break;
		case LIT:
			new_node = lit_to_cnf(node, witness);
			break;
		default:
			lerr(1, "to_cnf():\n\tunknown node type");
	};

	return new_node;
}

/* convert an AND node to CNF */
lnode *
and_to_cnf(lnode *src, lnode *witness)
{
	lnode		*new_witness = NULL;
	lnode		*next_expr, *ret, *clause1, *clause2, *clause3;

	/* if we have non-leaf nodes, we will need a new witness */
	if ((!is_leaf(src->left)) || (!is_leaf(src->right))) {
		new_witness = new_witness_node();
	}

	/* break down futher using an implication which holds */
	if (!is_leaf(src->left)) {

		next_expr = new_and_node(new_witness, src->right);
		ret = new_and_node(to_cnf(src->left, new_witness),
		    to_cnf(next_expr, witness));

	} else if (!is_leaf(src->right)) {

		next_expr = new_and_node(src->left, new_witness);
		ret = new_and_node(to_cnf(src->right, new_witness),
		    to_cnf(next_expr, witness));

	} else {
		/* so this is an implication in 3-address form, break down
		 * X<=>A.B   becomes   (A+~X).(B+~X).(~A+~B+X)
		 */
		clause1 = new_or_node(src->left, new_not_node(witness));
		clause2 = new_or_node(src->right, new_not_node(witness));
		clause3 = new_or_node(witness,
		    new_or_node(new_not_node(src->left),
		    new_not_node(src->right)));

		ret = new_and_node(clause1, new_and_node(clause2, clause3));
	}

	return ret;
}

/* convert an OR node to CNF */
lnode *
or_to_cnf(lnode *src, lnode *witness)
{
	lnode		*new_witness = NULL;
	lnode		*next_expr, *ret, *clause1, *clause2, *clause3;

	/* if we have non-leaf nodes, we will need a new witness */
	if ((!is_leaf(src->left)) || (!is_leaf(src->right))) {
		new_witness = new_witness_node();
	}

	/* break down futher using an implication which holds */
	if (!is_leaf(src->left)) {

		next_expr = new_or_node(new_witness, src->right);
		ret = new_and_node(to_cnf(src->left, new_witness),
		    to_cnf(next_expr, witness));
		
	} else if (!is_leaf(src->right)) {

		next_expr = new_or_node(src->left, new_witness);
		ret = new_and_node(to_cnf(src->right, new_witness),
		    to_cnf(next_expr, witness));

	} else {
		/*
		 * so this is an implication in 3-address form, break down
		 * X<=>A+B   becomes   (~A+X).(~B+X).(A+B+~X)
		 */
		clause1 = new_or_node(new_not_node(src->left), witness);
		clause2 = new_or_node(new_not_node(src->right), witness);
		clause3 = new_or_node(src->left, new_or_node(src->right,
		   new_not_node(witness)));

		ret = new_and_node(clause1, new_and_node(clause2, clause3));
	}

	return ret;
}

/* convert a LIT node to CNF */
lnode *
lit_to_cnf(lnode *src, lnode *witness)
{
	lnode		*clause1 = new_or_node(new_not_node(src), witness);
	lnode		*clause2 = new_or_node(new_lit_node(src->lit->name),
	    new_not_node(witness)); /* again copy */

	return new_and_node(clause1, clause2);
}

/* convert a NOT node to CNF */
lnode *
not_to_cnf(lnode *src, lnode *witness)
{
	lnode		*ret, *new_witness, *clause1, *clause2, *transform;

	/* NOT can have no right hand operand */
	assert(src->right == NULL);

	if (is_leaf(src->left)) {

		clause1 = new_or_node(new_not_node(src->left),
		    new_not_node(witness));	
		clause2 = new_or_node(src->left, witness);
		ret = new_and_node(clause1, clause2);

	} else {
		new_witness = new_witness_node();
		transform = new_not_node(new_witness);

		ret = new_and_node(to_cnf(transform, witness),
		    to_cnf(src->left, new_witness)); 
	}

	return ret;
}

lnode *
start_cnf_conversion(lnode *start)
{
	lnode		*witness = new_witness_node();
	return new_and_node(witness, to_cnf(start, witness));
}
