/* ------------------
 * Logic2CNF
 * $Id: uq_elim.c 509 2010-06-11 10:29:55Z eb771 $
 *
 * Copyright (c) 2010, Edd Barrett (University of Kent)
 * <e.barrett@kent.ac.uk>
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "logic2cnf.h"
#include "y.tab.h"
#include "queue.h"

/* from lits.c */
extern RB_HEAD(cnf_tree, literal_)		literals;

/* from cnf_clauses.c */
extern TAILQ_HEAD(ncs_head, cnf_clause_)	cnf_clauses;

SLIST_HEAD(uqe_head, uq_elim_lit_)		uq_elims;
unsigned long int				num_uq_elims = 0;

void
add_uq_elim(char *lit_name)
{
	uq_elim_lit		*ulit;

	ulit = (uq_elim_lit *) xmalloc(sizeof(uq_elim_lit));
	ulit->lit = get_literal(lit_name);

	if (ulit->lit == NULL)
		lerr(1, "add_uq_elim():\n\tundefined literal '%s' in uq_elim");

	SLIST_INSERT_HEAD(&uq_elims, ulit, entries);
	num_uq_elims ++;

	/* turn on resolution, it will be needed */
	resolve_temps = 1;
}

void
remove_uq_lits_from_cnf_clauses()
{
	cnf_clause		*clause, *next_clause;
	uq_elim_lit		*uqe_lit;
	unsigned long int	done = 0;

	/* for every lit, we need to .... */
	SLIST_FOREACH(uqe_lit, &uq_elims, entries) {

		if (enable_progress_report) {
			fprintf(stderr, "\rProgress: UQ eliminating "
			    "lit %8lu/%8lu", ++done, num_uq_elims);
		}

		/*
		 * loop over cnf_clauses.
		 * could be deleted so use the unwind method incase
		 */
		for (clause = TAILQ_FIRST(&cnf_clauses);
		    clause != TAILQ_END(&cnf_clauses);
		    clause = next_clause) {
			next_clause = TAILQ_NEXT(clause, entries);
			remove_uq_lits_from_clause(uqe_lit, clause);
		}

		/*
		 * now this literal does not exist in node clauses
		 * we may aswell remove it.
		 * XXX is this correct? Always gone?
		 */
		RB_REMOVE(cnf_tree, &literals, uqe_lit->lit);
		free_literal(uqe_lit->lit);
		num_literals --;
	}

	if (enable_progress_report)
		fprintf(stderr, "\n");

	if (l_debug) {
		fprintf(stderr, "Logic after UQ elimination: \n");
		all_cnf_clauses_to_text(1);
	}
}

void
remove_uq_lits_from_clause(uq_elim_lit *ulit, cnf_clause *clause)
{
	cnf_clause_lit		*ncl, *next_ncl;

	for (ncl = TAILQ_FIRST(&clause->lits);
	    ncl != TAILQ_END(&clause->lits);
	    ncl = next_ncl) {
		next_ncl = TAILQ_NEXT(ncl, entries);

		/* remove lit from clause if it matches */
		if (ncl->lit == ulit->lit) {
			TAILQ_REMOVE(&clause->lits, ncl, entries);
			clause->len --;
			free(ncl);
		}
	}

	/* if that clause is empty now, remove it */
	if (clause->len < 1) {
		TAILQ_REMOVE(&cnf_clauses, clause, entries);
		num_cnf_clauses --;
		free(clause);
	}
}

void
free_uq_elims()
{
	uq_elim_lit		*lit;

	while (!SLIST_EMPTY(&uq_elims)) {
		lit = SLIST_FIRST(&uq_elims);
		SLIST_REMOVE_HEAD(&uq_elims, entries);
		free(lit);
	}
}
